import java.io.IOException;
import java.util.Arrays;
import java.util.List;

import ucar.nc2.dataset.NetcdfDataset;
import ucar.nc2.dt.grid.GridDataset;
import ucar.nc2.dt.GridDatatype;
import ucar.nc2.dt.GridCoordSystem;
import ucar.ma2.Range;
import ucar.ma2.Array;
import ucar.ma2.Index;
import ucar.ma2.DataType;

public class NCTest4 {

  // Holds various properties of the variable.
  private static class VariableProperties {
    public int[] dims = new int[2];
    public int[] chunkSizes;
    public int bytesPerValue;
    public long connectionTime;
  } // VariableProperties class

  // Creates and populates the properties for a variable.
  private static VariableProperties getProperties (
    String datasetName,
    String varName
  ) {

    long startTime = System.currentTimeMillis();

    var props = new VariableProperties();
    
    try (NetcdfDataset ncDataset = NetcdfDataset.openDataset (datasetName, true, null)) {

      GridDataset gridDataset = new GridDataset (ncDataset);
      List<GridDatatype> gridList = gridDataset.getGrids();
      GridDatatype grid = gridDataset.findGridDatatype (varName);

      int width = grid.getXDimension().getLength();
      int height = grid.getYDimension().getLength();
      props.dims[0] = width;
      props.dims[1] = height;

      var chunkAtt = grid.findAttributeIgnoreCase ("_ChunkSizes");
      if (chunkAtt != null) {
        props.chunkSizes = (int[]) chunkAtt.getValues().copyTo1DJavaArray();
      } // if

      var type = grid.getVariable().getOriginalDataType();
      switch (type) {
      case BYTE:
      case CHAR:
        props.bytesPerValue = 1;
        break;
      case SHORT:
        props.bytesPerValue = 2;
        break;
      case INT:
      case FLOAT:
        props.bytesPerValue = 4;
        break;
      case DOUBLE:
      case LONG:
        props.bytesPerValue = 8;
        break;
      default:
        throw new IllegalStateException ("Unknown data type: " + type);
      } // switch

    } // try
    catch (Exception e) { throw new RuntimeException (e); }

    long endTime = System.currentTimeMillis();
    props.connectionTime = (endTime - startTime);

    return (props);

  } // getProperties

  private static class Tile {
    public int size;
    public int minX;
    public int maxX;
    public int minY;
    public int maxY;
    public String toString () {
      return ("Tile[size=" + size + ",minX=" + minX + ",maxX=" + maxX + ",minY=" + minY + ",maxY=" + maxY + "]");
    } // toString
  } // VariableProperties class

  // Gets the extents for a random tile of the specified size contained within
  // the given dimensions.
  private static Tile randomTile (
    int[] dims,
    int tileSize
  ) {
  
    Tile tile = new Tile();
    tile.size = tileSize;
    
    tile.minX = (int) (Math.random()*(dims[0]-1-(tileSize-1)));
    tile.maxX = tile.minX + (tileSize-1);

    tile.minY = (int) (Math.random()*(dims[1]-1-(tileSize-1)));
    tile.maxY = tile.minY + (tileSize-1);

    return (tile);
    
  } // randomTile

  // Holds the result values for a test of the server.
  private static class TestResults {
    public long time;
    public int count;
  } // TimeResults

  /**
   * Tests the connection to a NetCDF dataset.
   *
   * @param datasetName the full URL to the dataset.
   * @param varName the variable name within the dataset to access.
   * @param tile the tile to use for testing.
   * @param stride the stride to use for testing.
   *
   * @return the operation time in milliseconds.
   */
  private static TestResults testConnection (
    String datasetName,
    String varName,
    Tile tile,
    int stride
  ) {

    var results = new TestResults();

    long startTime = System.currentTimeMillis();

    int timeIndex = 0;
    int levelIndex = 0;
    
    try (NetcdfDataset ncDataset = NetcdfDataset.openDataset (datasetName, true, null)) {

      GridDataset gridDataset = new GridDataset (ncDataset);
      List<GridDatatype> gridList = gridDataset.getGrids();
      GridDatatype grid = gridDataset.findGridDatatype (varName);

      GridDatatype subset = grid.makeSubset (
        new Range (0, 0),  // runtime
        new Range (0, 0),  // ensemble
        new Range (timeIndex, timeIndex),
        new Range (levelIndex, levelIndex),
        new Range (tile.minY, tile.maxY, stride),
        new Range (tile.minX, tile.maxX, stride)
      );

      Array dataArray = subset.readDataSlice (0, 0, -1, -1);
      Index valueIndex = dataArray.getIndex();
      var shape = valueIndex.getShape();
      results.count = shape[0]*shape[1];

//      System.out.println ("Retrieved data with shape = " + Arrays.toString (shape));

    } // try
    catch (Exception e) { throw new RuntimeException (e); }

    long endTime = System.currentTimeMillis();
    results.time = endTime - startTime;

    return (results);

  } // testConnection

  // Runs the main test program.
  public static void main (String[] argv) throws IOException {

    String datasetName = argv[0];
    String varName = argv[1];

    var props = getProperties (datasetName, varName);
    System.out.println (datasetName + "." + varName + " [");
    System.out.println ("  dims = " + Arrays.toString (props.dims));
    System.out.println ("  chunkSizes = " + Arrays.toString (props.chunkSizes));
    System.out.println ("  bytesPerValue = " + props.bytesPerValue);
    System.out.println ("  connectionTime = " + props.connectionTime + " ms");
    System.out.println ("]");

    int n = 20;
    int tests = 0;
    double avgRate = 0;
    double avgNt = 0;
    System.out.println ("Performing " + n + " random tile download tests");
    for (int i = 0; i < n; i++) {
    
      int minDim = Math.min (props.dims[0], props.dims[1]);
      // Size is between 20% and 50% of the smaller dimension
      int tileSize = (minDim/5) + ((int) (Math.random()*(minDim/2 - minDim/5)));

      Tile tile = randomTile (props.dims, tileSize);

      var r2 = testConnection (datasetName, varName, tile, 1);
//      System.out.println ("Download time with full tile = " + r2.time + " ms");
      long t2 = r2.time;
      int n2 = r2.count;

      var r1 = testConnection (datasetName, varName, tile, 2);
//      System.out.println ("Download time with strided tile = " + r1.time + " ms");
      long t1 = r1.time;
      int n1 = r1.count;

//      if (t2 < t1) System.out.println ("Invalid test, T2 < T1");

      if (t2 > t1) {

        long dt = t2 - t1;
        double rate = (dt/1000.0)/((n2 - n1)*props.bytesPerValue);
        avgRate += rate;
        double nt = (t1 - rate*n1) / 1000;
        avgNt += nt;
        tests++;

        System.out.println ("[" + (i+1) + "]" +
          " tile=" + tile +
          " nt=" + String.format ("%.3f", nt) + "s" +
          " rate=" + String.format ("%.3f", (1/rate)/1024) + "kb/s"
        );

      } // if
      else {
        System.out.println ("[" + (i+1) + "]" +
          " tile=" + tile +
          " test failed, t1 >= t2"
        );
      } // else

    } // for
    
    avgNt /= tests;
    avgRate /= tests;
    System.out.println ("Avg network request time nt = " + String.format ("%.3f", avgNt) + "s");
    System.out.println ("Avg network download rate = " + String.format ("%.3f", (1/avgRate)/1024) + "kb/s");

  } // main
} // NCTest4 class
